package explorviz.live_trace_processing.filter.reconstruction;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.concurrent.ConcurrentSkipListMap;
import java.util.concurrent.TimeUnit;

import explorviz.live_trace_processing.filter.AbstractFilter;
import explorviz.live_trace_processing.filter.IPipeReceiver;
import explorviz.live_trace_processing.filter.reduction.summarization.TracePatternSummarizationFilter;
import explorviz.live_trace_processing.reader.TimeProvider;
import explorviz.live_trace_processing.record.IRecord;
import explorviz.live_trace_processing.record.event.AbstractOperationEventRecord;
import explorviz.live_trace_processing.record.misc.TerminateRecord;
import explorviz.live_trace_processing.record.misc.TimedPeriodRecord;
import explorviz.live_trace_processing.record.trace.Trace;

public final class TraceReconstructionFilter extends AbstractFilter {
	private static final int RINGBUFFER_LENGTH = 32;
	private static final int OUTPUT_BATCH_SIZE = 256;

	private final long maxTraceTimeout;

	private final Map<Long, TraceReconstructionBuffer> traceId2trace = new ConcurrentSkipListMap<Long, TraceReconstructionBuffer>();

	public TraceReconstructionFilter(final long maxTraceTimeout, final IPipeReceiver sinkReceiver) {
		super(
				new TracePatternSummarizationFilter(TimeUnit.MILLISECONDS.toNanos(990),
						sinkReceiver), RINGBUFFER_LENGTH, OUTPUT_BATCH_SIZE,
				"Reconstructed traces/sec");
		this.maxTraceTimeout = maxTraceTimeout;
	}

	@Override
	public void processRecord(final IRecord record) {
		if (record instanceof AbstractOperationEventRecord) {
			final AbstractOperationEventRecord abstractOperationEvent = ((AbstractOperationEventRecord) record);

			final long traceId = abstractOperationEvent.getTraceId();
			final TraceReconstructionBuffer traceBuffer = getBufferForTraceId(abstractOperationEvent
					.getTraceId());
			traceBuffer.insertEvent(abstractOperationEvent);

			if (traceBuffer.isFinished()) {
				traceId2trace.remove(traceId);
				sendOutValidTrace(traceBuffer.toTrace());
			}
		} else if (record instanceof Trace) {
			final Trace trace = (Trace) record;
			if (trace.isValid()) {
				deliver(trace);
			} else {
				getBufferForTraceId(trace.getTraceEvents().get(0).getTraceId());
			}
		} else if (record instanceof TimedPeriodRecord) {
			checkForTimeouts(TimeProvider.getCurrentTimestamp());
			periodicFlush(record);
		} else if (record instanceof TerminateRecord) {
			terminate();
		} else {
			deliver(record);
		}
	}

	private TraceReconstructionBuffer getBufferForTraceId(final long traceId) {
		TraceReconstructionBuffer traceBuffer = traceId2trace.get(traceId);
		if (traceBuffer == null) {
			traceBuffer = new TraceReconstructionBuffer();
			traceId2trace.put(traceId, traceBuffer);
		}
		return traceBuffer;
	}

	private void checkForTimeouts(final long timestamp) {
		final long traceTimeout = timestamp - maxTraceTimeout;
		final List<Long> traceIdsToRemove = new ArrayList<Long>();

		for (final Entry<Long, TraceReconstructionBuffer> entry : traceId2trace.entrySet()) {
			final TraceReconstructionBuffer traceBuffer = entry.getValue();
			if (traceBuffer.getMaxLoggingTimestamp() <= traceTimeout) {
				sendOutInvalidTrace(traceBuffer.toTrace());
				traceIdsToRemove.add(entry.getKey());
			}
		}

		for (final Long traceIdToRemove : traceIdsToRemove) {
			traceId2trace.remove(traceIdToRemove);
		}
	}

	private void sendOutValidTrace(final Trace trace) {
		trace.setValid(true);
		deliver(trace);
	}

	private void sendOutInvalidTrace(final Trace trace) {
		deliver(trace);
	}

	private void terminate() {
		for (final TraceReconstructionBuffer entry : traceId2trace.values()) {
			sendOutInvalidTrace(entry.toTrace());
		}
		traceId2trace.clear();
	}
}